import os
import sys
from datetime import datetime
import logging
import json
import io
from warcio import ArchiveIterator
from warcio.utils import BUFF_SIZE
import duckdb
import pyarrow as pa
from io import BytesIO

from lxml import etree, html

READ_SIZE = BUFF_SIZE * 4

THRESHOLD = 250

def bufcount(filename):
    """Count number of lines"""
    f = open(filename)                  
    lines = 0
    buf_size = 1024 * 1024
    read_f = f.read # loop optimization

    buf = read_f(buf_size)
    while buf:
        lines += buf.count('\n')
        buf = read_f(buf_size)
    f.close() 
    return lines

def cdx_size_counter(filename):
    return bufcount(filename) - 1


DEFAULT_LINK_ATTRS = ['class', 'id', 'href']

class Indexer:
    """Indexes WARC file metadata"""

    def __init__(self):
        pass

    def index_content(self, fromfile, tofile='warcindex.db'):
        """Generates SQLite database as WARC index"""
        from rich.progress import track
        from rich import print

        parser = etree.HTMLParser()  

        logging.debug("Indexing %s" % fromfile)
        resp = open(fromfile, "rb")  
        iterator = ArchiveIterator(resp)
        
        con = duckdb.connect(tofile)     
 
        cdx_filename = fromfile.rsplit('.', 2)[0] + '.cdx'
        records_num = None
        if os.path.exists(cdx_filename):
            records_num = cdx_size_counter(cdx_filename)
            print('CDX file found. Estimated number of WARC records %d' % (records_num))
        else:
            print("No CDX file. Cant measure progress")
        n = 0 
        list_records = []
        list_headers = []
        list_links = []
        for record in iterator:
            if record.rec_type != "response":
                continue

            n += 1      
            if records_num is not None:
                if n % THRESHOLD == 0: print('Processed %d (%0.2f%%) records' % (n, n*100.0 / records_num))            
            else:
                if n % THRESHOLD == 0: print('Processed %d records' % (n))
            if record.http_headers is not None:
                dbrec = {}
                dbrec['warc_id'] = record.rec_headers.get_header("WARC-Record-ID").rsplit(':', 1)[-1].strip('>')
                dbrec['url'] = record.rec_headers.get_header("WARC-Target-URI")
                content_type = record.http_headers.get_header("content-type")                
                dbrec['content_type'] = content_type
                charset = None
                content_type_no_ch = content_type
                if content_type is not None and content_type.find(';') > -1:
                    content_type_no_ch, charset = content_type.split(';', 1)
                    content_type_no_ch = content_type_no_ch.strip().lower()
                    if charset.find('=') > -1:
                         charset = charset.split('=', 1)[-1].lower().strip()

                if content_type_no_ch == 'text/html':
                    stream = record.content_stream()                                        
                    out_raw = BytesIO()
                    buf = stream.read(READ_SIZE)
                    while buf:
                        out_raw.write(buf)
                        buf = stream.read(READ_SIZE)
                    try:
                        root = etree.fromstring(out_raw.getvalue(), parser)  
                        links = root.xpath('//a')
                        for l in links:
                            lrec = {'warc_id' : dbrec['warc_id'], 'url' : dbrec['url'], '_text' : l.text}
                            for att in DEFAULT_LINK_ATTRS:
                                if att in l.attrib.keys(): lrec[att] = l.attrib[att]
                            list_links.append(lrec)
                    except KeyboardInterrupt:
                        pass

                dbrec['c_type'] = content_type_no_ch
                dbrec['c_type_charset'] = charset
                dbrec['offset'] = iterator.get_record_offset()
                dbrec['length'] = iterator.get_record_length()
                warc_date  = record.rec_headers.get_header("WARC-Date")
                dbrec['rec_date'] = datetime.strptime(warc_date, "%Y-%m-%dT%H:%M:%S%z")
                dbrec['content_length'] = int(record.rec_headers.get_header("Content-Length"))
                dbrec['status_code'] = int(record.http_headers.get_statuscode())                
                dbrec['source'] = fromfile
                dbrec['filename'] = dbrec['url'].rsplit("?", 1)[0].rsplit("/", 1)[-1].lower() 
                dbrec['ext'] = dbrec['filename'].rsplit(".", 1)[-1] if dbrec['filename'].find(".") > -1 else ""
                properties = []
                for key, value in dict(record.http_headers.headers).items():
                    properties.append({'key' : key, 'value' : 'value', 'warc_id' : dbrec['warc_id']})
                list_records.append(dbrec)
                list_headers.extend(properties)        

        resp.close()

        pa_records = pa.Table.from_pylist(list_records)
        pa_headers = pa.Table.from_pylist(list_headers)
        pa_links = pa.Table.from_pylist(list_links)
        con.sql("CREATE TABLE records AS SELECT * FROM pa_records")
        con.sql("CREATE TABLE headers AS SELECT * FROM pa_headers")
        con.sql("CREATE TABLE links AS SELECT * FROM pa_links")
        print(con.sql('select * from links'))

if __name__ == "__main__":
    indexer = Indexer()
    indexer.index_content(sys.argv[1])

